@@ -3,14 +3,14 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2006-2009, 2011-2012 Lukáš Lalinský
-# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014, 2018-2021 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2011-2012 Johannes Weißl
 # Copyright (C) 2011-2014 Michael Wiencek
 # Copyright (C) 2011-2014 Wieland Hoffmann
 # Copyright (C) 2013 Calvin Walton
-# Copyright (C) 2013-2014, 2017-2019 Laurent Monin
-# Copyright (C) 2013-2015, 2017 Sophist-UK
+# Copyright (C) 2013-2014, 2017-2021 Laurent Monin
+# Copyright (C) 2013-2015, 2017, 2021 Sophist-UK
 # Copyright (C) 2015 Frederik “Freso” S. Olesen
 # Copyright (C) 2016 Christoph Reiter
 # Copyright (C) 2016-2018 Sambhav Kothari
@@ -32,7 +32,8 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 
-from collections import defaultdict
+from collections import Counter
+from enum import IntEnum
 import re
 from urllib.parse import urlparse
 
@@ -49,6 +50,7 @@ from picard.coverart.image import (
     CoverArtImageError,
     TagCoverArtImage,
 )
+from picard.coverart.utils import types_from_id3
 from picard.file import File
 from picard.formats.mutagenext import (
     compatid3,
@@ -62,24 +64,22 @@ from picard.util import (
 from picard.util.tags import parse_comment_tag
 
 
+UNSUPPORTED_TAGS = {'r128_album_gain', 'r128_track_gain'}
+
 id3.GRP1 = compatid3.GRP1
 
-__IMAGE_TYPES = [
-    ("obi", 0),
-    ("tray", 0),
-    ("spine", 0),
-    ("sticker", 0),
-    ("other", 0),
-    ("front", 3),
-    ("back", 4),
-    ("booklet", 5),
-    ("track", 6),
-    ("medium", 6),
-]
 
-__ID3_IMAGE_TYPE_MAP = dict(__IMAGE_TYPES)
+class Id3Encoding(IntEnum):
+    LATIN1 = 0
+    UTF16 = 1
+    UTF16BE = 2
+    UTF8 = 3
 
-__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __IMAGE_TYPES])
+    def from_config(id3v2_encoding):
+        return {
+            'utf-8': Id3Encoding.UTF8,
+            'utf-16': Id3Encoding.UTF16
+        }.get(id3v2_encoding, Id3Encoding.LATIN1)
 
 
 def id3text(text, encoding):
@@ -87,23 +87,11 @@ def id3text(text, encoding):
     be encododed with the given numeric id3 encoding.
     """
 
-    if encoding == 0:
+    if encoding == Id3Encoding.LATIN1:
         return text.encode("latin1", "replace").decode("latin1")
     return text
 
 
-def image_type_from_id3_num(id3type):
-    return __ID3_REVERSE_IMAGE_TYPE_MAP.get(id3type, "other")
-
-
-def image_type_as_id3_num(texttype):
-    return __ID3_IMAGE_TYPE_MAP.get(texttype, 0)
-
-
-def types_from_id3(id3type):
-    return [image_type_from_id3_num(id3type)]
-
-
 def _remove_people_with_role(tags, frames, role):
     for frame in tags.values():
         if frame.FrameID in frames:
@@ -166,7 +154,7 @@ class ID3File(File):
         'TSO2': 'albumartistsort',
         'MVNM': 'movement'
     }
-    __rtranslate = dict([(v, k) for k, v in __translate.items()])
+    __rtranslate = {v: k for k, v in __translate.items()}
     __translate['GRP1'] = 'grouping'  # Always read, but writing depends on itunes_compatible_grouping
 
     __translate_freetext = {
@@ -198,7 +186,7 @@ class ID3File(File):
         'Writer': 'writer',
         'SHOWMOVEMENT': 'showmovement',
     }
-    __rtranslate_freetext = dict([(v, k) for k, v in __translate_freetext.items()])
+    __rtranslate_freetext = {v: k for k, v in __translate_freetext.items()}
     __translate_freetext['writer'] = 'writer'  # For backward compatibility of case
 
     # Freetext fields that are loaded case-insensitive
@@ -211,14 +199,14 @@ class ID3File(File):
         'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
         'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
     }
-    __translate_freetext_ci = dict([(b.lower(), a) for a, b in __rtranslate_freetext_ci.items()])
+    __translate_freetext_ci = {b.lower(): a for a, b in __rtranslate_freetext_ci.items()}
 
     # Obsolete tag names which will still be loaded, but will get renamed on saving
     __rename_freetext = {
         'Artists': 'ARTISTS',
         'Work': 'WORK',
     }
-    __rrename_freetext = dict([(v, k) for k, v in __rename_freetext.items()])
+    __rrename_freetext = {v: k for k, v in __rename_freetext.items()}
 
     _tipl_roles = {
         'engineer': 'engineer',
@@ -227,7 +215,7 @@ class ID3File(File):
         'DJ-mix': 'djmixer',
         'mix': 'mixer',
     }
-    _rtipl_roles = dict([(v, k) for k, v in _tipl_roles.items()])
+    _rtipl_roles = {v: k for k, v in _tipl_roles.items()}
 
     __other_supported_tags = ("discnumber", "tracknumber",
                               "totaldiscs", "totaltracks",
@@ -259,7 +247,7 @@ class ID3File(File):
         tags = file.tags or {}
         config = get_config()
         itunes_compatible = config.setting['itunes_compatible_grouping']
-        rating_user_email = config.setting['rating_user_email']
+        rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
         rating_steps = config.setting['rating_steps']
         # upgrade custom 2.3 frames to 2.4
         for old, new in self.__upgrade.items():
@@ -271,7 +259,7 @@ class ID3File(File):
             frameid = frame.FrameID
             if frameid in self.__translate:
                 name = self.__translate[frameid]
-                if frameid.startswith('T') or frameid in ["GRP1", "MVNM"]:
+                if frameid.startswith('T') or frameid in {"GRP1", "MVNM"}:
                     for text in frame.text:
                         if text:
                             metadata.add(name, text)
@@ -292,8 +280,12 @@ class ID3File(File):
                         metadata.add(name, text)
             elif frameid == "TMCL":
                 for role, name in frame.people:
-                    if role or name:
+                    if role == 'performer':
+                        role = ''
+                    if role:
                         metadata.add('performer:%s' % role, name)
+                    else:
+                        metadata.add('performer', name)
             elif frameid == "TIPL":
                 # If file is ID3v2.3, TIPL tag could contain TMCL
                 # so we will test for TMCL values and add to TIPL if not TMCL
@@ -301,7 +293,12 @@ class ID3File(File):
                     if role in self._tipl_roles and name:
                         metadata.add(self._tipl_roles[role], name)
                     else:
-                        metadata.add('performer:%s' % role, name)
+                        if role == 'performer':
+                            role = ''
+                        if role:
+                            metadata.add('performer:%s' % role, name)
+                        else:
+                            metadata.add('performer', name)
             elif frameid == 'TXXX':
                 name = frame.desc
                 name_lower = name.lower()
@@ -349,6 +346,7 @@ class ID3File(File):
                         comment=frame.desc,
                         support_types=True,
                         data=frame.data,
+                        id3_type=frame.type,
                     )
                 except CoverArtImageError as e:
                     log.error('Cannot load image from %r: %s' % (filename, e))
@@ -374,39 +372,42 @@ class ID3File(File):
         tags = self._get_tags(filename)
         config = get_config()
         if config.setting['clear_existing_tags']:
+            cover = tags.getall('APIC') if config.setting["preserve_images"] else None
             tags.clear()
+            if cover:
+                tags.setall('APIC', cover)
         images_to_save = list(metadata.images.to_be_saved_to_tags())
         if images_to_save:
             tags.delall('APIC')
 
-        encoding = {'utf-8': 3, 'utf-16': 1}.get(config.setting['id3v2_encoding'], 0)
+        encoding = Id3Encoding.from_config(config.setting['id3v2_encoding'])
 
         if 'tracknumber' in metadata:
             if 'totaltracks' in metadata:
                 text = '%s/%s' % (metadata['tracknumber'], metadata['totaltracks'])
             else:
                 text = metadata['tracknumber']
-            tags.add(id3.TRCK(encoding=0, text=id3text(text, 0)))
+            tags.add(id3.TRCK(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))
 
         if 'discnumber' in metadata:
             if 'totaldiscs' in metadata:
                 text = '%s/%s' % (metadata['discnumber'], metadata['totaldiscs'])
             else:
                 text = metadata['discnumber']
-            tags.add(id3.TPOS(encoding=0, text=id3text(text, 0)))
+            tags.add(id3.TPOS(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))
 
         if 'movementnumber' in metadata:
             if 'movementtotal' in metadata:
                 text = '%s/%s' % (metadata['movementnumber'], metadata['movementtotal'])
             else:
                 text = metadata['movementnumber']
-            tags.add(id3.MVIN(encoding=0, text=id3text(text, 0)))
+            tags.add(id3.MVIN(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))
 
         # This is necessary because mutagens HashKey for APIC frames only
         # includes the FrameID (APIC) and description - it's basically
         # impossible to save two images, even of different types, without
         # any description.
-        counters = defaultdict(lambda: 0)
+        counters = Counter()
         for image in images_to_save:
             desc = desctag = image.comment
             if counters[desc] > 0:
@@ -415,10 +416,10 @@ class ID3File(File):
                 else:
                     desctag = "(%i)" % counters[desc]
             counters[desc] += 1
-            tags.add(id3.APIC(encoding=0,
+            tags.add(id3.APIC(encoding=Id3Encoding.LATIN1,
                               mime=image.mimetype,
-                              type=image_type_as_id3_num(image.maintype),
-                              desc=id3text(desctag, 0),
+                              type=image.id3_type,
+                              desc=id3text(desctag, Id3Encoding.LATIN1),
                               data=image.data))
 
         tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
@@ -430,8 +431,11 @@ class ID3File(File):
 
             if not self.supports_tag(name):
                 continue
-            elif name.startswith('performer:'):
-                role = name.split(':', 1)[1]
+            elif name == 'performer' or name.startswith('performer:'):
+                if ':' in name:
+                    role = name.split(':', 1)[1]
+                else:
+                    role = 'performer'
                 for value in values:
                     if config.setting['write_id3v23']:
                         # TIPL will be upgraded to IPLS
@@ -442,7 +446,7 @@ class ID3File(File):
                 (lang, desc) = parse_comment_tag(name)
                 if desc.lower()[:4] == 'itun':
                     tags.delall('COMM:' + desc)
-                    tags.add(id3.COMM(encoding=0, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
+                    tags.add(id3.COMM(encoding=Id3Encoding.LATIN1, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
                 else:
                     tags.add(id3.COMM(encoding=encoding, desc=desc, lang=lang, text=values))
             elif name.startswith('lyrics:') or name == 'lyrics':
@@ -458,9 +462,10 @@ class ID3File(File):
             elif name == 'musicbrainz_recordingid':
                 tags.add(id3.UFID(owner='http://musicbrainz.org', data=bytes(values[0], 'ascii')))
             elif name == '~rating':
+                rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
                 # Search for an existing POPM frame to get the current playcount
                 for frame in tags.values():
-                    if frame.FrameID == 'POPM' and frame.email == config.setting['rating_user_email']:
+                    if frame.FrameID == 'POPM' and frame.email == rating_user_email:
                         count = getattr(frame, 'count', 0)
                         break
                 else:
@@ -468,7 +473,7 @@ class ID3File(File):
 
                 # Convert rating to range between 0 and 255
                 rating = int(round(float(values[0]) * 255 / (config.setting['rating_steps'] - 1)))
-                tags.add(id3.POPM(email=config.setting['rating_user_email'], rating=rating, count=count))
+                tags.add(id3.POPM(email=rating_user_email, rating=rating, count=count))
             elif name == 'grouping':
                 if config.setting['itunes_compatible_grouping']:
                     tags.add(id3.GRP1(encoding=encoding, text=values))
@@ -481,7 +486,7 @@ class ID3File(File):
             elif name in self.__rtranslate:
                 frameid = self.__rtranslate[name]
                 if frameid.startswith('W'):
-                    valid_urls = all([all(urlparse(v)[:2]) for v in values])
+                    valid_urls = all(all(urlparse(v)[:2]) for v in values)
                     if frameid == 'WCOP':
                         # Only add WCOP if there is only one license URL, otherwise use TXXX:LICENSE
                         if len(values) > 1 or not valid_urls:
@@ -581,9 +586,9 @@ class ID3File(File):
                     tags.delall(real_name)
                     tags.delall('TXXX:' + self.__rtranslate_freetext[name])
                 elif real_name == 'POPM':
-                    user_email = config.setting['rating_user_email']
+                    rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
                     for key, frame in list(tags.items()):
-                        if frame.FrameID == 'POPM' and frame.email == user_email:
+                        if frame.FrameID == 'POPM' and frame.email == rating_user_email:
                             del tags[key]
                 elif real_name in self.__translate:
                     tags.delall(real_name)
@@ -605,9 +610,8 @@ class ID3File(File):
 
     @classmethod
     def supports_tag(cls, name):
-        unsupported_tags = ['r128_album_gain', 'r128_track_gain']
-        return ((name and not name.startswith("~") and name not in unsupported_tags)
-                or name in ("~rating", "~length")
+        return ((name and not name.startswith("~") and name not in UNSUPPORTED_TAGS)
+                or name == "~rating"
                 or name.startswith("~id3"))
 
     def _get_tag_name(self, name):
@@ -650,36 +654,30 @@ class ID3File(File):
             tags.update_to_v24()
             tags.save(filename, v2_version=4, v1=v1)
 
-    @property
-    def new_metadata(self):
-        config = get_config()
-        if not config.setting["write_id3v23"]:
-            return self.metadata
+    def format_specific_metadata(self, metadata, tag, settings=None):
+        if not settings:
+            settings = get_config().setting
 
-        copy = Metadata()
-        copy.copy(self.metadata)
+        if not settings["write_id3v23"]:
+            return super().format_specific_metadata(metadata, tag, settings)
 
-        join_with = config.setting["id3v23_join_with"]
-        copy.multi_valued_joiner = join_with
+        values = metadata.getall(tag)
+        if not values:
+            return values
 
-        for name, values in copy.rawitems():
-            # ID3v23 can only save TDOR dates in YYYY format. Mutagen cannot
-            # handle ID3v23 dates which are YYYY-MM rather than YYYY or
-            # YYYY-MM-DD.
-            if name == "originaldate":
-                values = [v[:4] for v in values]
-            elif name == "date":
-                values = [(v[:4] if len(v) < 10 else v) for v in values]
+        if tag == "originaldate":
+            values = [v[:4] for v in values]
+        elif tag == "date":
+            values = [(v[:4] if len(v) < 10 else v) for v in values]
 
-            # If this is a multi-valued field, then it needs to be flattened,
-            # unless it's TIPL or TMCL which can still be multi-valued.
-            if (len(values) > 1 and name not in ID3File._rtipl_roles
-                    and not name.startswith("performer:")):
-                values = [join_with.join(values)]
+        # If this is a multi-valued field, then it needs to be flattened,
+        # unless it's TIPL or TMCL which can still be multi-valued.
+        if (len(values) > 1 and tag not in ID3File._rtipl_roles
+                and not tag.startswith("performer:")):
+            join_with = settings["id3v23_join_with"]
+            values = [join_with.join(values)]
 
-            copy[name] = values
-
-        return copy
+        return values
 
 
 class MP3File(ID3File):
