@@ -6,7 +6,7 @@
 # Copyright (C) 2007 Robert Kaye
 # Copyright (C) 2008 Gary van der Merwe
 # Copyright (C) 2008 Hendrik van Antwerpen
-# Copyright (C) 2008-2011, 2014-2015, 2018-2021 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014-2015, 2018-2022 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2009 Nikolai Prokoschenko
 # Copyright (C) 2011 Tim Blechmann
@@ -14,14 +14,17 @@
 # Copyright (C) 2011-2013 Michael Wiencek
 # Copyright (C) 2012 Your Name
 # Copyright (C) 2012-2013 Wieland Hoffmann
-# Copyright (C) 2013-2014, 2016, 2018-2019 Laurent Monin
-# Copyright (C) 2013-2014, 2017 Sophist-UK
+# Copyright (C) 2013-2014, 2016, 2018-2021 Laurent Monin
+# Copyright (C) 2013-2014, 2017, 2020 Sophist-UK
 # Copyright (C) 2016 Rahul Raturi
 # Copyright (C) 2016 Simon Legner
 # Copyright (C) 2016 Suhas
 # Copyright (C) 2016-2017 Sambhav Kothari
 # Copyright (C) 2018 Vishal Choudhary
-# Copyright (C) 2020 Gabriel Ferreira
+# Copyright (C) 2020-2021 Gabriel Ferreira
+# Copyright (C) 2021 Bob Swift
+# Copyright (C) 2021 Louis Sautier
+# Copyright (C) 2021 Petit Minion
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -66,7 +69,10 @@ from picard.config import (
     Option,
     get_config,
 )
-from picard.file import File
+from picard.file import (
+    File,
+    FileErrorType,
+)
 from picard.plugin import ExtensionPoint
 from picard.track import (
     NonAlbumTrack,
@@ -74,9 +80,11 @@ from picard.track import (
 )
 from picard.util import (
     icontheme,
+    iter_files_from_objects,
     natsort,
     normpath,
     restore_method,
+    strxfrm,
 )
 
 from picard.ui.collectionmenu import CollectionMenu
@@ -146,9 +154,7 @@ def get_match_color(similarity, basecolor):
 
 class MainPanel(QtWidgets.QSplitter):
 
-    options = [
-        Option("persist", "splitter_state", QtCore.QByteArray()),
-    ]
+    options = []
 
     columns = [
         (N_('Title'), 'title'),
@@ -167,6 +173,7 @@ class MainPanel(QtWidgets.QSplitter):
         (N_('Fingerprint status'), '~fingerprint'),
         (N_('Date'), 'date'),
         (N_('Original Release Date'), 'originaldate'),
+        (N_('Cover'), 'covercount'),
     ]
 
     _column_indexes = {column[1]: i for i, column in enumerate(columns)}
@@ -208,7 +215,7 @@ class MainPanel(QtWidgets.QSplitter):
         TreeItem.base_color = self.palette().base().color()
         TreeItem.text_color = self.palette().text().color()
         TreeItem.text_color_secondary = self.palette() \
-            .brush(QtGui.QPalette.Disabled, QtGui.QPalette.Text).color()
+            .brush(QtGui.QPalette.ColorGroup.Disabled, QtGui.QPalette.ColorRole.Text).color()
         TrackItem.track_colors = defaultdict(lambda: TreeItem.text_color, {
             File.NORMAL: interface_colors.get_qcolor('entity_saved'),
             File.CHANGED: TreeItem.text_color,
@@ -233,19 +240,12 @@ class MainPanel(QtWidgets.QSplitter):
         tab_order(prev, after)
 
     def save_state(self):
-        config = get_config()
-        config.persist["splitter_state"] = self.saveState()
         for view in self._views:
             view.save_state()
 
-    @restore_method
-    def restore_state(self):
-        config = get_config()
-        self.restoreState(config.persist["splitter_state"])
-
     def create_icons(self):
         if hasattr(QtWidgets.QStyle, 'SP_DirIcon'):
-            ClusterItem.icon_dir = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
+            ClusterItem.icon_dir = self.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_DirIcon)
         else:
             ClusterItem.icon_dir = icontheme.lookup('folder', icontheme.ICON_SIZE_MENU)
         AlbumItem.icon_cd = icontheme.lookup('media-optical', icontheme.ICON_SIZE_MENU)
@@ -261,6 +261,8 @@ class MainPanel(QtWidgets.QSplitter):
         FileItem.icon_file = QtGui.QIcon(":/images/file.png")
         FileItem.icon_file_pending = QtGui.QIcon(":/images/file-pending.png")
         FileItem.icon_error = icontheme.lookup('dialog-error', icontheme.ICON_SIZE_MENU)
+        FileItem.icon_error_not_found = icontheme.lookup('error-not-found', icontheme.ICON_SIZE_MENU)
+        FileItem.icon_error_no_access = icontheme.lookup('error-no-access', icontheme.ICON_SIZE_MENU)
         FileItem.icon_saved = QtGui.QIcon(":/images/track-saved.png")
         FileItem.icon_fingerprint = icontheme.lookup('fingerprint', icontheme.ICON_SIZE_MENU)
         FileItem.icon_fingerprint_gray = icontheme.lookup('fingerprint-gray', icontheme.ICON_SIZE_MENU)
@@ -318,12 +320,6 @@ class MainPanel(QtWidgets.QSplitter):
         for view in self._views:
             view.setSortingEnabled(sort)
 
-    def collapse_clusters(self, collapse=True):
-        if collapse:
-            self._views[0].collapseAll()
-        else:
-            self._views[0].expandAll()
-
     def select_object(self, obj):
         item = obj.item
         for view in self._views:
@@ -333,12 +329,12 @@ class MainPanel(QtWidgets.QSplitter):
                 break
 
 
-def paint_fingerprint_icon(painter, rect, icon):
+def paint_column_icon(painter, rect, icon):
     if not icon:
         return
     size = COLUMN_ICON_SIZE
     padding_h = COLUMN_ICON_BORDER
-    padding_v = (rect.height() - size) / 2
+    padding_v = (rect.height() - size) // 2
     target_rect = QtCore.QRect(rect.x() + padding_h, rect.y() + padding_v, size, size)
     painter.drawPixmap(target_rect, icon.pixmap(size, size))
 
@@ -346,21 +342,21 @@ def paint_fingerprint_icon(painter, rect, icon):
 class ConfigurableColumnsHeader(TristateSortHeaderView):
 
     def __init__(self, parent=None):
-        super().__init__(QtCore.Qt.Horizontal, parent)
+        super().__init__(QtCore.Qt.Orientation.Horizontal, parent)
         self._visible_columns = set([0])
 
         # The following are settings applied to default headers
         # of QTreeView and QTreeWidget.
         self.setSectionsMovable(True)
         self.setStretchLastSection(True)
-        self.setDefaultAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
+        self.setDefaultAlignment(QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignVCenter)
         self.setSectionsClickable(False)
         self.sortIndicatorChanged.connect(self.on_sort_indicator_changed)
 
         # enable sorting, but don't actually use it by default
         # XXX it would be nice to be able to go to the 'no sort' mode, but the
         #     internal model that QTreeWidget uses doesn't support it
-        self.setSortIndicator(-1, QtCore.Qt.AscendingOrder)
+        self.setSortIndicator(-1, QtCore.Qt.SortOrder.AscendingOrder)
         self.setDefaultSectionSize(100)
 
     def show_column(self, column, show):
@@ -372,10 +368,10 @@ class ConfigurableColumnsHeader(TristateSortHeaderView):
                 self.resizeSection(column, self.defaultSectionSize())
             self._visible_columns.add(column)
             if column == MainPanel.FINGERPRINT_COLUMN:
-                self.setSectionResizeMode(column, QtWidgets.QHeaderView.Fixed)
+                self.setSectionResizeMode(column, QtWidgets.QHeaderView.ResizeMode.Fixed)
                 self.resizeSection(column, COLUMN_ICON_SIZE)
             else:
-                self.setSectionResizeMode(column, QtWidgets.QHeaderView.Interactive)
+                self.setSectionResizeMode(column, QtWidgets.QHeaderView.ResizeMode.Interactive)
         elif column in self._visible_columns:
             self._visible_columns.remove(column)
 
@@ -383,10 +379,6 @@ class ConfigurableColumnsHeader(TristateSortHeaderView):
         for i, column in enumerate(MainPanel.columns):
             self.show_column(i, i in columns)
 
-    @property
-    def visible_columns(self):
-        return self._visible_columns
-
     def contextMenuEvent(self, event):
         menu = QtWidgets.QMenu(self)
         parent = self.parent()
@@ -424,19 +416,19 @@ class ConfigurableColumnsHeader(TristateSortHeaderView):
             painter.save()
             super().paintSection(painter, rect, index)
             painter.restore()
-            paint_fingerprint_icon(painter, rect, FileItem.icon_fingerprint_gray)
+            paint_column_icon(painter, rect, FileItem.icon_fingerprint_gray)
         else:
             super().paintSection(painter, rect, index)
 
     def on_sort_indicator_changed(self, index, order):
         if index == MainPanel.FINGERPRINT_COLUMN:
-            self.setSortIndicator(-1, QtCore.Qt.AscendingOrder)
+            self.setSortIndicator(-1, QtCore.Qt.SortOrder.AscendingOrder)
 
     def lock(self, is_locked):
         super().lock(is_locked)
-        fingerprint_column_index = MainPanel.FINGERPRINT_COLUMN
-        if not self.is_locked and self.count() > fingerprint_column_index:
-            self.setSectionResizeMode(fingerprint_column_index, QtWidgets.QHeaderView.Fixed)
+        column_index = MainPanel.FINGERPRINT_COLUMN
+        if not self.is_locked and self.count() > column_index:
+            self.setSectionResizeMode(column_index, QtWidgets.QHeaderView.ResizeMode.Fixed)
 
 
 class BaseTreeView(QtWidgets.QTreeWidget):
@@ -455,7 +447,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
         self.setAcceptDrops(True)
         self.setDragEnabled(True)
         self.setDropIndicatorShown(True)
-        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
+        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
 
         self.setSortingEnabled(True)
 
@@ -498,6 +490,8 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             if can_view_info:
                 menu.addAction(self.window.view_info_action)
             menu.addAction(self.window.browser_lookup_action)
+            if self.window.submit_cluster_action:
+                menu.addAction(self.window.submit_cluster_action)
             menu.addSeparator()
             menu.addAction(self.window.autotag_action)
             menu.addAction(self.window.analyze_action)
@@ -518,6 +512,10 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             menu.addAction(self.window.play_file_action)
             menu.addAction(self.window.open_folder_action)
             menu.addAction(self.window.browser_lookup_action)
+            if self.window.submit_file_as_recording_action:
+                menu.addAction(self.window.submit_file_as_recording_action)
+            if self.window.submit_file_as_release_action:
+                menu.addAction(self.window.submit_file_as_release_action)
             menu.addSeparator()
             menu.addAction(self.window.autotag_action)
             menu.addAction(self.window.analyze_action)
@@ -545,11 +543,14 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             menu.addMenu(releases_menu)
             loading = releases_menu.addAction(_('Loading...'))
             loading.setDisabled(True)
+            action_more = releases_menu.addAction(_('Show &more details...'))
+            action_more.triggered.connect(self.window.album_other_versions_action.trigger)
             bottom_separator = True
 
             if len(self.selectedItems()) == 1 and obj.release_group:
                 def _add_other_versions():
                     releases_menu.removeAction(loading)
+                    releases_menu.removeAction(action_more)
                     heading = releases_menu.addAction(obj.release_group.version_headings)
                     heading.setDisabled(True)
                     font = heading.font()
@@ -596,6 +597,9 @@ class BaseTreeView(QtWidgets.QTreeWidget):
                     versions_count = len(versions)
                     if versions_count > 1:
                         releases_menu.setTitle(_("&Other versions (%d)") % versions_count)
+
+                    releases_menu.addSeparator()
+                    action = releases_menu.addAction(action_more)
                 if obj.release_group.loaded:
                     _add_other_versions()
                 else:
@@ -680,10 +684,10 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             header.update_visible_columns([0, 1, 2])
             for i, size in enumerate([250, 50, 100]):
                 header.resizeSection(i, size)
-            self.sortByColumn(-1, QtCore.Qt.AscendingOrder)
+            self.sortByColumn(-1, QtCore.Qt.SortOrder.AscendingOrder)
 
     def supportedDropActions(self):
-        return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction
+        return QtCore.Qt.DropAction.CopyAction | QtCore.Qt.DropAction.MoveAction
 
     def mimeTypes(self):
         """List of MIME types accepted by this view."""
@@ -691,7 +695,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
 
     def dragEnterEvent(self, event):
         if event.mimeData().hasUrls():
-            event.setDropAction(QtCore.Qt.CopyAction)
+            event.setDropAction(QtCore.Qt.DropAction.CopyAction)
             event.accept()
         else:
             event.acceptProposedAction()
@@ -707,7 +711,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             pixmap = QtGui.QPixmap(rectangle.width(), rectangle.height())
             self.viewport().render(pixmap, QtCore.QPoint(), QtGui.QRegion(rectangle))
             drag.setPixmap(pixmap)
-            drag.exec_(QtCore.Qt.MoveAction)
+            drag.exec_(QtCore.Qt.DropAction.MoveAction)
 
     def mimeData(self, items):
         """Return MIME data for specified items."""
@@ -726,7 +730,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             mimeData.setUrls(files)
         return mimeData
 
-    def scrollTo(self, index, scrolltype=QtWidgets.QAbstractItemView.EnsureVisible):
+    def scrollTo(self, index, scrolltype=QtWidgets.QAbstractItemView.ScrollHint.EnsureVisible):
         # QTreeView.scrollTo resets the horizontal scroll position to 0.
         # Reimplemented to maintain current horizontal scroll position.
         hscrollbar = self.horizontalScrollBar()
@@ -740,7 +744,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
         new_paths = []
         tagger = QtCore.QObject.tagger
         for url in urls:
-            log.debug("Dropped the URL: %r", url.toString(QtCore.QUrl.RemoveUserInfo))
+            log.debug("Dropped the URL: %r", url.toString(QtCore.QUrl.UrlFormattingOption.RemoveUserInfo))
             if url.scheme() == "file" or not url.scheme():
                 filename = normpath(url.toLocalFile().rstrip("\0"))
                 file = tagger.files.get(filename)
@@ -748,7 +752,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
                     files.append(file)
                 else:
                     new_paths.append(filename)
-            elif url.scheme() in ("http", "https"):
+            elif url.scheme() in {"http", "https"}:
                 file_lookup = tagger.get_file_lookup()
                 file_lookup.mbid_lookup(url.path(), browser_fallback=False)
         if files:
@@ -759,7 +763,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
     def dropEvent(self, event):
         # Dropping with Alt key pressed forces all dropped files being
         # assigned to the same track.
-        if event.keyboardModifiers() == QtCore.Qt.AltModifier:
+        if event.keyboardModifiers() == QtCore.Qt.KeyboardModifier.AltModifier:
             self._move_to_multi_tracks = False
         return QtWidgets.QTreeView.dropEvent(self, event)
 
@@ -785,10 +789,12 @@ class BaseTreeView(QtWidgets.QTreeWidget):
         # application/picard.album-list
         albums = data.data("application/picard.album-list")
         if albums:
-            albums = [self.tagger.load_album(id) for id in bytes(albums).decode().split("\n")]
+            album_ids = bytes(albums).decode().split("\n")
+            log.debug("Dropped albums = %r", album_ids)
+            files = iter_files_from_objects(self.tagger.load_album(id) for id in album_ids)
             # Use QTimer.singleShot to run expensive processing outside of the drop handler.
-            QtCore.QTimer.singleShot(0, partial(self.tagger.move_files,
-                self.tagger.get_files_from_objects(albums), target))
+            move_files = partial(self.tagger.move_files, list(files), target)
+            QtCore.QTimer.singleShot(0, move_files)
             handled = True
         self._move_to_multi_tracks = True  # Reset for next drop
         return handled
@@ -811,7 +817,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             cluster_item.add_files(cluster.files)
 
     def moveCursor(self, action, modifiers):
-        if action in (QtWidgets.QAbstractItemView.MoveUp, QtWidgets.QAbstractItemView.MoveDown):
+        if action in {QtWidgets.QAbstractItemView.CursorAction.MoveUp, QtWidgets.QAbstractItemView.CursorAction.MoveDown}:
             item = self.currentItem()
             if item and not item.isSelected():
                 self.setCurrentItem(item)
@@ -889,12 +895,12 @@ class TreeItem(QtWidgets.QTreeWidgetItem):
             obj.item = self
         self.sortable = sortable
         self._sortkeys = {}
-        for column in [
+        for column in (
             MainPanel.LENGTH_COLUMN,
             MainPanel.TRACKNUMBER_COLUMN,
             MainPanel.DISCNUMBER_COLUMN,
-        ]:
-            self.setTextAlignment(column, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
+        ):
+            self.setTextAlignment(column, QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter)
         self.setSizeHint(MainPanel.FINGERPRINT_COLUMN, ICON_SIZE)
 
     def setText(self, column, text):
@@ -916,9 +922,9 @@ class TreeItem(QtWidgets.QTreeWidgetItem):
         if column == MainPanel.LENGTH_COLUMN:
             sortkey = self.obj.metadata.length or 0
         elif column in MainPanel.NAT_SORT_COLUMNS:
-            sortkey = natsort.natkey(self.text(column).lower())
+            sortkey = natsort.natkey(self.text(column))
         else:
-            sortkey = self.text(column).lower()
+            sortkey = strxfrm(self.text(column))
         self._sortkeys[column] = sortkey
         return sortkey
 
@@ -1040,7 +1046,7 @@ class NatAlbumItem(AlbumItem):
         if not tree_widget:
             return True
         order = tree_widget.header().sortIndicatorOrder()
-        return order == QtCore.Qt.AscendingOrder
+        return order == QtCore.Qt.SortOrder.AscendingOrder
 
 
 class TrackItem(TreeItem):
@@ -1135,7 +1141,12 @@ class FileItem(TreeItem):
     @staticmethod
     def decide_file_icon(file):
         if file.state == File.ERROR:
-            return FileItem.icon_error
+            if file.error_type == FileErrorType.NOTFOUND:
+                return FileItem.icon_error_not_found
+            elif file.error_type == FileErrorType.NOACCESS:
+                return FileItem.icon_error_no_access
+            else:
+                return FileItem.icon_error
         elif isinstance(file.parent, Track):
             if file.state == File.NORMAL:
                 return FileItem.icon_saved
